/* xrdev/examples/infrared_display.c
 *
 * simple continuous display of current infrared values
 *
 *                             ** COPYRIGHT NOTICE **
 *
 * Copyright (c) 1999, Nomadic Technologies, Inc.
 *
 * The contents of this file is copyrighted by Nomadic Technologies, Inc, and
 * is protected by United States copyright laws.  Unauthorized copying is
 * prohibited.  All rights reserved.
 *
 */

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "Nclient.h"

#include "robot_arg.h"



/* flag, set by the signal trapper to exit the main loop */
BOOL ir_done;



/* ------------------------------------------------------------------------
 * Function:     Usage
 * Purpose:      print the command line options and exit
 * Arguments:    *program_name : argv[0]
 * ------------------------------------------------------------------------ */
static void Usage(const char *program_name)
{
  fprintf(stderr, "Usage: %s [--host host[:port]] [--robot_id N]\n",
	  program_name);
  exit(-1);
}



/* ------------------------------------------------------------------------
 * Function:     SignalTrap
 * Purpose:      catch process interrupts, so that we can exit the main
 *                 loop and clean up after ourselves
 * ------------------------------------------------------------------------ */
static void SignalTrap(int signal_num)
{
  ir_done = TRUE;
  return;
}
	


int main(int argc, char *argv[])
{
  char *sched_host;
  const char *argv0;
  unsigned short sched_port;
  int i, j, error;
  long robot_id;
  struct N_InfraredSet *set;
  struct N_RobotState *rs;

  argv0 = ARG_GetProgramName(argv[0]);

  /* get the robot ID and scheduler from the command line */
  sched_host = "localhost";
  sched_port = 7073;
  robot_id = 1;
  if (ARG_GetStandardArgs(&argc, &argv, &sched_host, &sched_port, &robot_id))
  {
    Usage(argv0);
  }

  /* initialize the client library and connect to our robot */
  if ((error = N_InitializeClient(sched_host, sched_port)) != N_NO_ERROR ||
      (error = N_ConnectRobot(robot_id)) != N_NO_ERROR)
  {
    fprintf(stderr, "%s: error connecting to %s:%d (%d).\n", argv0,
	    sched_host, sched_port, error);
    exit(-1);
  }

  /* get a pointer to the state structure */
  rs = N_GetRobotState(robot_id);

  /* set TimeStampActive for all infrared sets so that N_GetInfared will
   * return timestamped data */
  for (j = 0; j < rs->InfraredController.InfraredSetCount; j++)
  {
    rs->InfraredController.InfraredSet[j].TimeStampActive = TRUE;
  }

  /* loop until SIGINT */
  ir_done = FALSE;
  signal(SIGINT, SignalTrap);
  while (!ir_done)
  {
    /* fill rs->InfraredController and all substructures with the
     * current values (and timestamps) */
    if ((error = N_GetInfrared(robot_id)) != N_NO_ERROR)
    {
      printf("%s: retrieval of sonar values failed (%d).\n", argv0, error);
      exit(-1);
    }

    printf("---readings-----------------------------"
	   "---------------------------------------\n");
    for (j = 0; j < rs->InfraredController.InfraredSetCount; j++)
    {
      set = &(rs->InfraredController.InfraredSet[j]); 
      for (i = 0; i < set->InfraredCount; i++)
      {
        printf("%9ld ", set->Infrared[i].Reading);
      }
      printf("\n");
    }
    printf("---time stamps --------------------------"
	   "---------------------------------------\n");
    for (j = 0; j < rs->InfraredController.InfraredSetCount; j++)
    {
      set = &(rs->InfraredController.InfraredSet[j]); 
      for (i = 0; i < set->InfraredCount; i++)
      {
        printf("%9ld ", set->Infrared[i].TimeStamp);
      }
      printf("\n");
    }
    printf("\n");

    sleep(1);
  }
  
  N_DisconnectRobot(robot_id);
  return 0;
}
