/* xrdev/examples/ir_push_xr.c
 *
 * infrared-based obstacle avoidance
 *
 *                             ** COPYRIGHT NOTICE **
 *
 * Copyright (c) 1999, Nomadic Technologies, Inc.
 *
 * The contents of this file is copyrighted by Nomadic Technologies, Inc, and
 * is protected by United States copyright laws.  Unauthorized copying is
 * prohibited.  All rights reserved.
 *
 */

#include <math.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "Nclient.h"

#include "robot_arg.h"


/* this program demonstrates obstacle avoidance using infrareds.  The
 * instantaneous velocity is calculated as the average of weighted vectors
 * opposite each infrared reading.  This velocity is subjected to a speed
 * cap and a proportional damping constant before being sent to the motors. */


/* acceleration values */
#define XY_ACCEL           300
#define PSI_ACCEL          1100

/* the energy level at which we decide to move */
#define MOVE_THRESHOLD     15

/* the dampening constant: v(t) = v(t-1) + k(u(t) - v(t-1))  where v
 * is the velocity given to the motors and u is the velocity requested
 * by the directed ir energy summation */
#define DERIVATIVE_K       0.07 

/* the requested velocities are averaged and multiplied by this factor */
#define SPEED_K            1.5

/* speed caps, mm/sec */
#define MAX_VX             400
#define MAX_VY             400

/* set to output to stdout lines of the form: iteration_no ux uv uy vy 
 * this output can easily be run through a graphing program to observe the
 * effect of the control system */
#undef OUTPUT_VELOCITIES



/* flag, set by the signal trapper to exit the main loop */
static BOOL ir_done;



/* ------------------------------------------------------------------------
 * Function:     Usage
 * Purpose:      print the command line options and exit
 * Arguments:    *program_name : argv[0]
 * ------------------------------------------------------------------------ */
static void Usage(const char *program_name)
{
  fprintf(stderr, "Usage: %s [--host host[:port]] [--robot_id N]\n",
	  program_name);
  exit(-1);
}



/* ------------------------------------------------------------------------
 * Function:     SignalTrap
 * Purpose:      catch process interrupts, so that we can exit the main
 *                 loop and clean up after ourselves
 * ------------------------------------------------------------------------ */
static void SignalTrap(int signal_num)
{
  ir_done = TRUE;
  return;
}



int main(int argc, char *argv[])
{
  char *sched_host;
  const char *argv0;
  unsigned short sched_port;
  unsigned long iteration, hit_count;
  double heading, ux, uy, vx, vy, old_vx, old_vy;
  int i, j, k, error;
  long robot_id;
  struct N_RobotState *rs;
  struct N_Infrared *ir;

  argv0 = ARG_GetProgramName(argv[0]);

  /* get the robot ID and scheduler from the command line */
  sched_host = "localhost";
  sched_port = 7073;
  robot_id = 1;
  if (ARG_GetStandardArgs(&argc, &argv, &sched_host, &sched_port, &robot_id))
  {
    Usage(argv0);
  }

  /* initialize the client library and connect to our robot */
  if ((error = N_InitializeClient(sched_host, sched_port)) != N_NO_ERROR ||
      (error = N_ConnectRobot(robot_id)) != N_NO_ERROR)
  {
    fprintf(stderr, "%s: error connecting to %s:%d (%d).\n", argv0,
	    sched_host, sched_port, error);
    exit(-1);
  }

  /* should be no other command line arguments */
  if (argc > 1)
  {
    Usage(argv0);
  }

  rs = N_GetRobotState(robot_id);

  rs->AxisSet.Global = FALSE;
  rs->AxisSet.Axis[N_XTRANSLATION].Mode = N_AXIS_VELOCITY;
  rs->AxisSet.Axis[N_XTRANSLATION].Acceleration = XY_ACCEL;
  rs->AxisSet.Axis[N_YTRANSLATION].Mode = N_AXIS_VELOCITY;
  rs->AxisSet.Axis[N_YTRANSLATION].Acceleration = XY_ACCEL;
  rs->AxisSet.Axis[N_ROTATION].Mode = N_AXIS_STOP;
  rs->AxisSet.Axis[N_ROTATION].Update = TRUE;

  old_vx = old_vy = 0.0;
  iteration = 0;

  ir_done = FALSE;
  signal(SIGINT, SignalTrap);
  while (!ir_done)
  {
    ++iteration;

    /* reset */
    ux = uy = 0.0;
    hit_count = 0;

    /* get infrareds readings */
    if ((error = N_GetInfrared(robot_id)) != N_NO_ERROR)
    {
      fprintf(stderr, "%s: retrieval of infrared values failed (%d).\n", 
	      argv0, error);
      exit(-1);
    }

    heading = M_PI/6.0 + M_PI/24.0;

    /* j loops through three doors; i loops through 8 infrareds; k loops
     * through top and bottom infrared sets */
    for (j = 0; j < 3; ++j)
    {
      for (i = 0; i < 8; ++i)
      {
	for (k = 0; k < 2; ++k)
	{
	  ir = &rs->InfraredController.InfraredSet[j*2 + k].Infrared[i];

	  if (ir->Reading >= MOVE_THRESHOLD)
	  {
	    ++hit_count;

	    /* sin(pi+x) = -sin(x); cos(pi+x) = -cos(x) */
	    ux -= ir->Reading*cos(heading);
	    uy -= ir->Reading*sin(heading);
	  }
	}
	
	/* heading of the next infrared pair (top and bottom) */
	heading += M_PI/12.0;
      }
    }

    /* average and scale.  the averaging has the nice side effect that total
     * velocity goes down as more energy comes in. */
    if (hit_count > 1)
    {
      ux /= hit_count;
      uy /= hit_count;
    }
    ux *= SPEED_K;
    uy *= SPEED_K;

    /* derivative control */
    vx = old_vx + DERIVATIVE_K*(ux - old_vx);
    vy = old_vy + DERIVATIVE_K*(uy - old_vy);
    old_vx = vx;
    old_vy = vy;

    /* velocity cap */
    if (vx > MAX_VX)
    {
      vx = MAX_VX;
    }
    if (vy > MAX_VY)
    {
      vy = MAX_VY;
    }

    /* drive */
    rs->AxisSet.Axis[N_XTRANSLATION].DesiredSpeed = (long)vx;
    rs->AxisSet.Axis[N_XTRANSLATION].Update = TRUE;
    rs->AxisSet.Axis[N_YTRANSLATION].DesiredSpeed = (long)vy;
    rs->AxisSet.Axis[N_YTRANSLATION].Update = TRUE;

    if ((error = N_SetAxes(robot_id)) != N_NO_ERROR &&
	error != N_AXES_NOT_READY)
    {
      fprintf(stderr, "%s: failure while setting axes (%d).\n", argv0, error);
      exit(-1);
    }
  }

  rs->AxisSet.Axis[N_XTRANSLATION].Mode = N_AXIS_STOP;
  rs->AxisSet.Axis[N_XTRANSLATION].Update = TRUE;
  rs->AxisSet.Axis[N_YTRANSLATION].Mode = N_AXIS_STOP;
  rs->AxisSet.Axis[N_YTRANSLATION].Update = TRUE;
  rs->AxisSet.Axis[N_ROTATION].Mode = N_AXIS_STOP;
  rs->AxisSet.Axis[N_ROTATION].Update = TRUE;

  if ((error = N_SetAxes(robot_id)) != N_NO_ERROR &&
      error != N_AXES_NOT_READY)
  {
    fprintf(stderr, "%s: failure while stopping axes (%d).\n", argv0, error);
  }

  N_DisconnectRobot(robot_id);
  return 0;
}
