/* xrdev/examples/lift.c
 *
 * XR4000 lift mechanism demonstration program
 *
 *                             ** COPYRIGHT NOTICE **
 *
 * Copyright (c) 1999, Nomadic Technologies, Inc.
 *
 * The contents of this file is copyrighted by Nomadic Technologies, Inc, and
 * is protected by United States copyright laws.  Unauthorized copying is
 * prohibited.  All rights reserved.
 *
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "Nclient.h"

#include "robot_arg.h"



/* ------------------------------------------------------------------------
 * Function:     MoveLift
 * Purpose:      start a lift motion with the given mode and dest position
 * Arguments:    *lift : from the state structure
 *               axis : the axis number
 *               mode : the axis mode
 *               position : the desired destination position for this move
 *               robot_id : the robot id to send the command to
 * Return:       the results of N_SetLift
 * ------------------------------------------------------------------------ */
static int MoveLift(struct N_LiftController *lift, unsigned int axis,
		    char mode, long position, long robot_id)
{
  lift->Axis[axis].Mode = mode;
  lift->Axis[axis].DesiredPosition = position;
  lift->Axis[axis].Update = TRUE;

  return N_SetLift(robot_id);
}

  

int main(int argc, char *argv[])
{
  BOOL done;
  char *sched_host, string[50], mode;
  const char *argv0;
  unsigned short sched_port;
  int result, choice;
  unsigned int axis;
  long position, acc, force, velocity, robot_id;
  unsigned long time, t0;
  struct N_LiftController *as;

  argv0 = ARG_GetProgramName(argv[0]);

  /* get the robot ID and scheduler from the command line */
  sched_host = "localhost";
  sched_port = 7073;
  robot_id = 1;
  ARG_GetStandardArgs(&argc, &argv, &sched_host, &sched_port, &robot_id);

  if ((result = N_InitializeClient(sched_host, sched_port)) != N_NO_ERROR ||
      (result = N_ConnectRobot(robot_id)) != N_NO_ERROR)
  {
    fprintf(stderr, "%s: error connecting (%d).\n", argv0, result);
    exit(-1);
  }
  
  as = &N_GetRobotState(robot_id)->LiftController;
  
  printf("\nNote: the Lift Mechanism must be zeroed (Item 1) and deployed "
	 "(Item 2) before\naxis movement is allowed.\n\n");       

  done = FALSE;
  while (!done)
  {
    printf("Lift utility:\n 0. Quit\n 1. Zero\n 2. Deploy\n 3. Retract\n"
	   " 4. Set axis position\n 5. Velocity move\n 6. Change Max Force\n"
	   " 7. Change Acceleration\n 8. Read axes\n"
	   " 9. Pick up an object and then drop it.\n"
	   "Choice: ");
    fflush(stdout);

    scanf("%d", &choice);
    printf("\n");

    if ((result = N_GetLift(robot_id)) != N_NO_ERROR)
    {
      fprintf(stderr, "%s: error retrieving currnet lift values (%d).\n",
	      argv0, result);
      exit(-1);
    }

    switch (choice)
    {
    case 0: /* quit */

      done = TRUE;
      break;

    case 1: /* zero */

      printf("Force the operation [y/n]? ");
      fflush(stdout);
      scanf("%s", string);
      if ((result = N_ZeroLift(robot_id, string[0] == 'y' || string[0] == 'Y'))
	  != N_NO_ERROR)
      {
	fprintf(stderr, "Error zeroing (%d).\n", result);
	break;
      }
      while (1)
      {
	if ((result = N_GetLift(robot_id)) != N_NO_ERROR)
	{
	  fprintf(stderr, "Error retrieving lift state (%d)\n", result);
	  break;
	}
	if (!as->InProgress)
	{
	  break;
	}
	printf("d:%s gp: %-4ld gv:%-4ld  "
	       "lp:%-4ld lv:%-4ld\n", as->Deployed ? "y" : "n",
	       as->Axis[N_GRIP].Position, as->Axis[N_GRIP].Velocity,
	       as->Axis[N_LIFT].Position, as->Axis[N_LIFT].Velocity);
      }
      printf("\n");
      break;

    case 2: /* deploy */

      if ((result = N_DeployLift(robot_id)) != N_NO_ERROR)
      {
	fprintf(stderr, "Error deploying (%d).\n", result);
	break;
      }
      while (1)
      {
	if ((result = N_GetLift(robot_id)) != N_NO_ERROR)
	{
	  fprintf(stderr, "Error retrieving lift state (%d)\n", result);
	  break;
	}
	printf("d:%s  gp: %-4ld>%-4ld gv:%-4ld  "
	       "lp:%-4ld>%-4ld lv:%-4ld\n", as->Deployed ? "y" : "n",
	       as->Axis[N_GRIP].Position, as->Axis[N_GRIP].DesiredPosition,
	       as->Axis[N_GRIP].Velocity,
	       as->Axis[N_LIFT].Position, as->Axis[N_LIFT].DesiredPosition,
	       as->Axis[N_LIFT].Velocity);
	if (!as->InProgress)
	{
	  break;
	}
      }
      printf("\n");
      break;

    case 3: /* retract */

      if ((result = N_RetractLift(robot_id)) != N_NO_ERROR)
      {
	fprintf(stderr, "Error retracting (%d).\n", result);
	break;
      }
      while (1)
      {
	if ((result = N_GetLift(robot_id)) != N_NO_ERROR)
	{
	  fprintf(stderr, "Error retrieving lift state (%d)\n", result);
	  break;
	}
	printf("d:%s  gp: %-4ld>%-4ld gv:%-4ld  "
	       "lp:%-4ld>%-4ld lv:%-4ld\n", as->Deployed ? "y" : "n",
	       as->Axis[N_GRIP].Position, as->Axis[N_GRIP].DesiredPosition,
	       as->Axis[N_GRIP].Velocity,
	       as->Axis[N_LIFT].Position, as->Axis[N_LIFT].DesiredPosition,
	       as->Axis[N_LIFT].Velocity);
	if (!as->InProgress)
	{
	  break;
	}
      }
      printf("\n");
      break;

    case 4: /* set axis */
      
      printf("Grip or lift (g/l)? ");
      scanf("%s", string);
      axis = string[0] == 'g' ? N_GRIP : N_LIFT;
      as->Axis[axis].TimeStampActive = TRUE;
      as->Axis[axis].DesiredVelocity = 500;
      printf("Absolute or relative position move (a/r)? ");
      scanf("%s", string);
      if (string[0] == 'a')
      {
	mode = N_LIFT_POSITION_ABSOLUTE;
      }
      else
      {
	mode = N_LIFT_POSITION_RELATIVE;
      }
      printf("Position (0.1mm)? ");
      scanf("%ld", &position);
      if ((result = MoveLift(as, axis, mode, position, robot_id))
	  != N_NO_ERROR)
      {
	printf("Error setting the axis (%ld %ld).\n",
	       as->Axis[N_GRIP].Status, as->Axis[N_LIFT].Status);
      }
      else
      {
	while (1)
	{
	  if ((result = N_GetLift(robot_id)) != N_NO_ERROR)
	  {
	    fprintf(stderr, "Error retrieving lift state (%ld %ld)\n",
		    as->Axis[N_GRIP].Status, as->Axis[N_LIFT].Status);
	    break;
	  }
	  printf("pos:%ld->%ld   velocity:%-4ld  time: %lu\n",
		 as->Axis[axis].Position, as->Axis[axis].DesiredPosition,
		 as->Axis[axis].Velocity, as->Axis[axis].TimeStamp);
	  if (as->Axis[axis].Velocity == 0)
	  {
	    break;
	  }
	}
      }
      as->Axis[axis].TimeStampActive = FALSE;
      break;

    case 5: /* velocity move */

      printf("Grip or lift (g/l)? ");
      scanf("%s", string);
      axis = string[0] == 'g' ? N_GRIP : N_LIFT;
      printf("Velocity (0.1mm/s)? ");
      scanf("%ld", &velocity);
      printf("Time? ");
      scanf("%lu", &time);
      as->Axis[axis].Mode = N_AXIS_VELOCITY;
      as->Axis[axis].DesiredVelocity = velocity;
      as->Axis[axis].Update = TRUE;
      as->Axis[axis].TimeStampActive = TRUE;
      if ((result = N_SetLift(robot_id)) != N_NO_ERROR)
      {
	printf("Error setting the axis (%ld %ld).\n",
	       as->Axis[N_GRIP].Status, as->Axis[N_LIFT].Status);
      }
      else
      {
	if ((result = N_GetLift(robot_id)) != N_NO_ERROR)
	{
	  fprintf(stderr, "Error retrieving lift state (%ld %ld)\n",
		  as->Axis[N_GRIP].Status, as->Axis[N_LIFT].Status);
	  break;
	}
	t0 = as->Axis[axis].TimeStamp;
	printf("t0 is %lu\n", t0);
	while (1)
	{
	  if ((result = N_GetLift(robot_id)) != N_NO_ERROR)
	  {
	    fprintf(stderr, "Error retrieving lift state (%ld %ld)\n",
		    as->Axis[N_GRIP].Status, as->Axis[N_LIFT].Status);
	    break;
	  }
	  printf("pos:%ld->%ld   velocity:%-4ld  time: %lu\n",
		 as->Axis[axis].Position, as->Axis[axis].DesiredPosition,
		 as->Axis[axis].Velocity, as->Axis[axis].TimeStamp);
	  if (as->Axis[axis].Velocity == 0 ||
	      as->Axis[axis].TimeStamp - t0 >= time)
	  {
	    break;
	  }
	}
      }
      as->Axis[axis].Mode = N_AXIS_VELOCITY;
      as->Axis[axis].DesiredVelocity = 0;
      as->Axis[axis].Update = TRUE;
      if ((result = N_SetLift(1)) != N_NO_ERROR)
      {
	printf("Error setting the axis (%ld %ld).\n",
	       as->Axis[N_GRIP].Status, as->Axis[N_LIFT].Status);
      }
      else
      {
	while (1)
	{
	  if ((result = N_GetLift(robot_id)) != N_NO_ERROR)
	  {
	    fprintf(stderr, "Error retrieving lift state (%ld %ld)\n",
		    as->Axis[N_GRIP].Status, as->Axis[N_LIFT].Status);
	    break;
	  }
	  printf("pos:%ld->%ld   velocity:%-4ld  time: %lu\n",
		 as->Axis[axis].Position, as->Axis[axis].DesiredPosition,
		 as->Axis[axis].Velocity, as->Axis[axis].TimeStamp);
	  if (as->Axis[axis].Velocity == 0)
	  {
	    break;
	  }
	}
      }
      as->Axis[axis].TimeStampActive = FALSE;
      printf("\n");
      break;

    case 6: /* set max force */

      printf("Grip or lift (g/l)? ");
      scanf("%s", string);
      axis = string[0] == 'g' ? N_GRIP : N_LIFT;
      printf("Force (0-100)? ");
      scanf("%ld", &force);
      as->Axis[axis].MaxMotor=force;
      N_SetLift(robot_id);
      printf("Force set to %ld\n", force);
      break;

    case 7: /* set acceleration */
      printf("Grip or lift (g/l)? ");
      scanf("%s", string);
      axis = string[0] == 'g' ? N_GRIP : N_LIFT;
      printf("Acceleration (0.1mm/s/s)? ");
      scanf("%ld", &acc);
      as->Axis[axis].Acceleration=acc;
      N_SetLift(robot_id);
      printf("Acceleration set to %ld\n", acc);
      break;

    case 8: /* read axes */

      for (axis = 0; axis < as->AxisCount; ++axis)
      {
	as->Axis[axis].TimeStampActive = TRUE;
      }
      if ((result = N_GetLift(robot_id)) != N_NO_ERROR)
      {
	fprintf(stderr, "Error retrieving lift state (%d)\n\n", result);
      }
      else
      {
	printf("deployed: %s\n", as->Deployed ? "y" : "n");
	for (axis = 0; axis < as->AxisCount; ++axis)
	{
	  printf("axis %d  position: %ld>%ld  velocity:%ld  maxmotor:%ld  "
		 "timestamp: %lu\n", axis,
		 as->Axis[axis].Position, as->Axis[axis].DesiredPosition,
		 as->Axis[axis].Velocity, as->Axis[axis].MaxMotor,
		 as->Axis[axis].TimeStamp);
	}
      }
      for (axis = 0; axis < as->AxisCount; ++axis)
      {
	as->Axis[axis].TimeStampActive = FALSE;
      }
      break;

    case 9: /* pick up an object... */

      /* set velocities */
      as->Axis[N_GRIP].DesiredVelocity = 500;
      as->Axis[N_LIFT].DesiredVelocity = 500;
      printf("Opening the gripper and lowering the lift...\n");
      if ((result = MoveLift(as, N_GRIP, N_LIFT_POSITION_ABSOLUTE, 1300,
			     robot_id)) != N_NO_ERROR)
      {
	printf("Error setting the axis (%d).\n", result);
	break;	
      }
      if ((result = MoveLift(as, N_LIFT, N_LIFT_POSITION_ABSOLUTE,
			     -4800, robot_id)) != N_NO_ERROR)
      {
	printf("Error setting the axis (%d).\n", result);
	break;
      }
      while (1)
      {
	if ((result = N_GetLift(robot_id)) != N_NO_ERROR)
	{
	  fprintf(stderr, "Error retrieving lift state (%d)\n", result);
	  break;
	}
	if (as->Axis[N_LIFT].Velocity == 0 && as->Axis[N_GRIP].Velocity == 0)
	{
	  break;
	}
      }
      printf("Closing the gripper...\n");
      /* set maximum motor to low value so we don't crush the can */
      as->Axis[N_GRIP].MaxMotor=30;
      if ((result = MoveLift(as, N_GRIP, N_LIFT_POSITION_ABSOLUTE, 0,
			     robot_id)) != N_NO_ERROR)
      {
	printf("Error setting the axis (%d).\n", result);
	break;
      }
      while (1)
      {
	if ((result = N_GetLift(robot_id)) != N_NO_ERROR)
	{
	  fprintf(stderr, "Error retrieving lift state (%d)\n", result);
	  break;
	}
	/* if gripper errors, we have picked up something! */
	if (as->Axis[N_GRIP].Status != 0)
	{
	  printf("Picked up.\n");
	  break;
	}
	/* if the gripper closes, we have not pickup up something */
	if (as->Axis[N_GRIP].Velocity == 0)
	{
	  printf("Could not pick up.\n");
	  break;
	}
      }

      printf("Lifting...\n");
      if ((result = MoveLift(as, N_LIFT, N_LIFT_POSITION_ABSOLUTE,
			     0, robot_id)) != N_NO_ERROR)
      {
	printf("Error setting the axis (%d).\n", result);
	break;
      }
      while (1)
      {
	if ((result = N_GetLift(robot_id)) != N_NO_ERROR)
	{
	  fprintf(stderr, "Error retrieving lift state (%d)\n", result);
	  break;
	}
	if (as->Axis[N_LIFT].Velocity == 0)
	{
	  break;
	}
      }
      /* reset maximum motor to maximum value */
      as->Axis[N_GRIP].MaxMotor=100;
      printf("Opening the gripper...\n");
      if ((result = MoveLift(as, N_GRIP, N_LIFT_POSITION_ABSOLUTE,
			     2000, robot_id)) !=
	  N_NO_ERROR)
      {
	printf("Error setting the axis (%d).\n", result);
	break;
      }
      while (1)
      {
	if ((result = N_GetLift(robot_id)) != N_NO_ERROR)
	{
	  fprintf(stderr, "Error retrieving lift state (%d).\n", result);
	  break;
	}
	if (as->Axis[N_GRIP].Velocity == 0)
	{
	  break;
	}
      }
      break;

    default:

      printf("Invalid choice.\n\n");
      break;
    }
  } /* loop until the user selects option #0 */

  N_DisconnectRobot(robot_id);
  return 0;
}
