/* xrdev/examples/sick_display.c
 *
 * simple continuous display of current SICK laser values
 *
 *                             ** COPYRIGHT NOTICE **
 *
 * Copyright (c) 1999, Nomadic Technologies, Inc.
 *
 * The contents of this file is copyrighted by Nomadic Technologies, Inc, and
 * is protected by United States copyright laws.  Unauthorized copying is
 * prohibited.  All rights reserved.
 *
 */

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "Nclient.h"

#include "robot_arg.h"



#define NUM_POINTS           361      /* number of points to retrieve */



/* flag, set by the signal trapper to exit the main loop */
BOOL sick_done;



/* ------------------------------------------------------------------------
 * Function:     Usage
 * Purpose:      print the command line options and exit
 * Arguments:    *program_name : argv[0]
 * ------------------------------------------------------------------------ */
static void Usage(const char *program_name)
{
  fprintf(stderr, "Usage: %s [--host host[:port]] [--robot_id N]\n",
          program_name);
  exit(-1);
}



/* ------------------------------------------------------------------------
 * Function:     SignalTrap
 * Purpose:      catch process interrupts, so that we can exit the main
 *                 loop and clean up after ourselves
 * ------------------------------------------------------------------------ */
static void SignalTrap(int signal_num)
{
  sick_done = TRUE;
  return;
}
	


int main(int argc, char *argv[])
{
  char *sched_host;
  const char *argv0;
  unsigned short sched_port;
  int error;
  long robot_id;
  unsigned int i, j;
  struct N_RobotState *rs;

  argv0 = ARG_GetProgramName(argv[0]);

  /* get the robot ID and scheduler from the command line */
  sched_host = "localhost";
  sched_port = 7073;
  robot_id = 1;
  if (ARG_GetStandardArgs(&argc, &argv, &sched_host, &sched_port, &robot_id))
  {
    Usage(argv0);
  }

  /* check for other command line options */
  if (argc > 1)
  {
    Usage(argv0);
  }

  /* initialize the client library and connect to our robot */
  if ((error = N_InitializeClient(sched_host, sched_port)) != N_NO_ERROR ||
      (error = N_ConnectRobot(robot_id)) != N_NO_ERROR)
  {
    fprintf(stderr, "%s: error connecting to %s:%d (%d).\n", argv0,
	    sched_host, sched_port, error);
    exit(-1);
  }

  rs = N_GetRobotState(robot_id);

  if (rs->S550Set.S550Count == 0)
  {
    fprintf(stderr, "%s: this robot is not serving data for any "
	    "SICK lasers.\n", argv0);
    exit(-1);
  }

  /* configure each SICK to return the desired number of points and
   * timestamp all data */
  for (i = 0; i < rs->S550Set.S550Count; ++i)
  {
    rs->S550Set.S550[i].RequestedPoints = NUM_POINTS;
    rs->S550Set.S550[i].TimeStampActive = TRUE;
  }

  /* loop until SIGINT */
  sick_done = FALSE;
  signal(SIGINT, SignalTrap);
  while (!sick_done)
  {
    /* fill in rs->S550Set and all substructures with the current values */
    if ((error = N_GetS550(robot_id)) != N_NO_ERROR)
    {
      fprintf(stderr, "%s: error retrieving S550 values (%d).\n",
	      argv0, error);
      exit(-1);
    }
    
    for (i = 0; i < rs->S550Set.S550Count; ++i)
    {
      printf("---%2d: readings (timestamp is %9ld)--------"
	     "-----------------------------\n", 
	     i, rs->S550Set.S550[i].TimeStamp);
      for (j = 0; j < rs->S550Set.S550[i].RequestedPoints; ++j)
      {
	printf("%5ld ", rs->S550Set.S550[i].Readings[j]);
	if (j % 13 == 12)
	{
	  printf("\n");
	}
      }
      if (j % 13 != 12)
      {
	printf("\n");
      }
    }
    printf("\n");

    sleep(1);
  }
    
  N_DisconnectRobot(robot_id);
  return 0;
}
