/* xrdev/examples/sonar_conf.c
 *
 * utility to manage sonar firing
 *
 *                             ** COPYRIGHT NOTICE **
 *
 * Copyright (c) 1999, Nomadic Technologies, Inc.
 *
 * The contents of this file is copyrighted by Nomadic Technologies, Inc, and
 * is protected by United States copyright laws.  Unauthorized copying is
 * prohibited.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Nclient.h"

#include "robot_arg.h"



/* ------------------------------------------------------------------------
 * Function:     Usage
 * Purpose:      print the command line options and exit
 * Arguments:    *program_name : argv[0]
 * ------------------------------------------------------------------------ */
static void Usage(const char *program_name)
{
  fprintf(stderr, "Usage: %s [options] [[(all|set_no) "
	  "(full|off|fo1 fo2 fo3 ... end)]...]\n"
	  "Options:\n"
	  " --host host[:port]      specify the host to connect to\n"
	  " --robot_id N            specify the robot ID number\n"
	  " --help                  display this message\n", program_name);
  exit(-1);
}



/* ------------------------------------------------------------------------
 * Function:     SetFiringOrder
 * Purpose:      set the firing order for the given set
 * Arguments:    *set : the sonar set to work with
 *               unsigned int *fo : the firing order string; NULL means full
 * ------------------------------------------------------------------------ */
static void SetFiringOrder(struct N_SonarSet *set, unsigned int *fo)
{
  int i;

  if (fo == NULL)
  {
    for (i = 0; i < set->SonarCount; ++i)
    {
      set->FiringOrder[i] = i;
    }
  }
  else
  {
    for (i = 0; i < set->SonarCount && *fo != N_END_SONAR_FIRING_ORDER; ++i)
    {
      set->FiringOrder[i] = fo[i];
    }
  }
  set->FiringOrder[i] = N_END_SONAR_FIRING_ORDER;
  return;
}



int main(int argc, char *argv[])
{
  char *sched_host, *argp;
  const char *argv0;
  unsigned short sched_port;
  unsigned int *fo, *fop;
  int i, set, sn, error, min_scnt, max_scnt, min_scnt_set;
  long robot_id;
  struct N_SonarController *sc;

  argv0 = ARG_GetProgramName(argv[0]);

  /* get the robot ID and scheduler from the command line */
  sched_host = "localhost";
  sched_port = 7073;
  robot_id = 1;
  if (ARG_GetStandardArgs(&argc, &argv, &sched_host, &sched_port, &robot_id))
  {
    Usage(argv0);
  }

  /* initialize the client library and connect to our robot */
  if ((error = N_InitializeClient(sched_host, sched_port)) != N_NO_ERROR ||
      (error = N_ConnectRobot(robot_id)) != N_NO_ERROR)
  {
    fprintf(stderr, "%s: error connecting to %s:%d (%d).\n", argv0,
	    sched_host, sched_port, error);
    exit(-1);
  }

  /* get a pointer to the sonar controller */
  sc = &N_GetRobotState(robot_id)->SonarController;
  
  /* get the current sonar configuration */
  if ((error = N_GetSonarConfiguration(robot_id)) != N_NO_ERROR)
  {
    fprintf(stderr, "%s: error retrieving sonar configuration (%d).\n",
	    argv0, error);
    exit(-1);
  }

  /* find the set with the min. number of sonars */
  min_scnt = N_MAX_SONAR_COUNT;
  max_scnt = 0;
  for (i = 0; i < sc->SonarSetCount; ++i)
  {
    if (sc->SonarSet[i].SonarCount < min_scnt)
    {
      min_scnt_set = i;
      min_scnt = sc->SonarSet[i].SonarCount;
    }
    if (sc->SonarSet[i].SonarCount > max_scnt)
    {
      max_scnt = sc->SonarSet[i].SonarCount;
    }
  }

  fo = (unsigned int *)malloc(sizeof(unsigned int)*(max_scnt + 1));

  /* loop, setting sonar firing orders */
  i = 1;
  while (i < argc)
  {
    /* first, get the set num */
    if (strcasecmp(argv[i], "all") == 0)
    {
      set = N_MAX_SONAR_SET_COUNT;
    }
    else
    {
      set = strtol(argv[i], &argp, 10);
      if (*argp != '\0' || set >= sc->SonarSetCount)
      {
	fprintf(stderr, "%s: \"%s\" is an invalid set number.\n",
		argv[0], argv[i]);
	free(fo);
	Usage(argv[0]);
      }
    }

    ++i;
    if (i > argc)
    {
      break;
    }

    if (strcasecmp(argv[i], "on") == 0 ||
	strcasecmp(argv[i], "full") == 0)
    {
      fop = NULL;
      ++i;
    }
    else
    {
      fop = fo;
      while (i < argc)
      {
	if (strcasecmp(argv[i], "done") ==0 ||
	    strcasecmp(argv[i], "end") == 0)
	{
	  ++i;
	  break;
	}
	else if (strcasecmp(argv[i], "off") == 0 ||
		 strcasecmp(argv[i], "empty") == 0)
	{
	  fop = fo;
	  ++i;
	  break;
	}

	/* get the next sonar in the firing order */
	sn = strtol(argv[i], &argp, 10);
	if (*argp != '\0')
	{
	  fprintf(stderr, "%s: \"%s\" is an invalid sonar number.\n",
		  argv[0], argv[i]);
	  free(fo);
	  Usage(argv[0]);
	}
	if (set == N_MAX_SONAR_SET_COUNT)
	{
	  if (sn >= min_scnt)
	  {
	    fprintf(stderr, "%s: set %d only has %d sonars; %d exceeds this "
		    "minimum.\n", argv[0], min_scnt_set, min_scnt, sn);
	    free(fo);
	    Usage(argv[0]);
	  }
	}
	else if (sn >= sc->SonarSet[set].SonarCount)
	{
	  fprintf(stderr, "%s: set %d only has %d sonars; %d is invalid.\n",
		  argv[0], set, sc->SonarSet[set].SonarCount, sn);
	  free(fo);
	  Usage(argv[0]);
	}
	*fop++ = sn;
	++i;
      } /* end of loop setting oridnal firing order */
      *fop = N_END_SONAR_FIRING_ORDER;
      fop = fo;
    } /* not explicitly set to "full" */

    /* set the firing order for this set */
    if (set == N_MAX_SONAR_SET_COUNT)
    {
      for (set = 0; set < sc->SonarSetCount; ++set)
      {
	SetFiringOrder(&sc->SonarSet[set], fop);
      }
    }
    else
    {
      SetFiringOrder(&sc->SonarSet[set], fop);
    }
  }

  free(fo);

  /* set the new sonar configuration */
  if ((error = N_SetSonarConfiguration(robot_id)) != N_NO_ERROR)
  {
    fprintf(stderr, "%s: error setting sonar configuration (%d).\n",
	    argv0, error);
    exit(-1);
  }

  /* re-read, to double check */
  if ((error = N_GetSonarConfiguration(robot_id)) != N_NO_ERROR)
  {
    fprintf(stderr, "%s: error retrieving sonar configuration (%d).\n",
	    argv0, error);
    exit(-1);
  }

  for (set = 0; set < sc->SonarSetCount; ++set)
  {
    printf("%d: ", set);
    if (sc->SonarSet[set].FiringOrder[0] == N_END_SONAR_FIRING_ORDER)
    {
      printf("off\n");
    }
    else
    {
      i = 0;
      while (i < sc->SonarSet[set].SonarCount &&
	     sc->SonarSet[set].FiringOrder[i] != N_END_SONAR_FIRING_ORDER)
      {
	printf("%d ", sc->SonarSet[set].FiringOrder[i]);
	++i;
      }
      printf("\n");
    }
  }

  N_DisconnectRobot(robot_id);
  return 0;
}
