/* xrdev/examples/sonar_display.c
 *
 * simple continuous display of current sonar values
 *
 *                             ** COPYRIGHT NOTICE **
 *
 * Copyright (c) 1999, Nomadic Technologies, Inc.
 *
 * The contents of this file is copyrighted by Nomadic Technologies, Inc, and
 * is protected by United States copyright laws.  Unauthorized copying is
 * prohibited.  All rights reserved.
 *
 */

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>

#include "Nclient.h"

#include "robot_arg.h"



/* flag, set by the signal trapper to exit the main loop */
static BOOL sonar_done;



/* ------------------------------------------------------------------------
 * Function:     Usage
 * Purpose:      print the command line options and exit
 * Arguments:    *program_name : argv[0]
 * ------------------------------------------------------------------------ */
static void Usage(const char *program_name)
{
  fprintf(stderr, "Usage: %s [--host host[:port]] [--robot_id N]\n",
	  program_name);
  exit(-1);
}



/* ------------------------------------------------------------------------
 * Function:     SignalTrap
 * Purpose:      catch process interrupts, so that we can exit the main
 *                 loop and clean up after ourselves
 * ------------------------------------------------------------------------ */
static void SignalTrap(int signal_num)
{
  sonar_done = TRUE;
  return;
}



int main(int argc, char *argv[])
{
  char *sched_host;
  const char *argv0;
  unsigned short sched_port;
  int i, j, error;
  unsigned int old_firing_orders[N_MAX_SONAR_COUNT + 1][N_MAX_SONAR_SET_COUNT];
  long robot_id;
  long old_firing_delays[N_MAX_SONAR_SET_COUNT];
  struct N_SonarSet *set;
  struct N_RobotState *rs;

  argv0 = ARG_GetProgramName(argv[0]);

  /* get the robot ID and scheduler from the command line */
  sched_host = "localhost";
  sched_port = 7073;
  robot_id = 1;
  if (ARG_GetStandardArgs(&argc, &argv, &sched_host, &sched_port, &robot_id))
  {
    Usage(argv0);
  }

  /* initialize the client library and connect to our robot */
  if ((error = N_InitializeClient(sched_host, sched_port)) != N_NO_ERROR ||
      (error = N_ConnectRobot(robot_id)) != N_NO_ERROR)
  {
    fprintf(stderr, "%s: error connecting to %s:%d (%d).\n", argv0,
	    sched_host, sched_port, error);
    exit(-1);
  }

  /* get a pointer to the state structure */
  rs = N_GetRobotState(robot_id);

  /* get the current sonar configuration so we can store it for later */
  if ((error = N_GetSonarConfiguration(robot_id)) != N_NO_ERROR)
  {
    printf("%s: retrieval of current configuration failed (%d).\n",
	   argv0, error);
    exit(-1);
  }

  for (j = 0; j < rs->SonarController.SonarSetCount; j++)
  {
    set = &rs->SonarController.SonarSet[j]; 

    /* record the original firing order and firing delay */
    memcpy(old_firing_orders[j], set->FiringOrder, N_MAX_SONAR_COUNT + 1);
    old_firing_delays[j] = set->FiringDelay;

    /* configure the sonars for full firing order with no delay */
    set->FiringDelay = 0;
    set->TimeStampActive = TRUE;
    for (i=0; i < set->SonarCount; i++)
    {
      set->FiringOrder[i] = i;
    }
  }
  if ((error = N_SetSonarConfiguration(robot_id)) != N_NO_ERROR)
  {
    printf("%s: sonar configuration failed (%d).\n", argv0, error);
    exit(-1);
  }

  /* loop until SIGINT */
  sonar_done = FALSE;
  signal(SIGINT, SignalTrap);
  while (!sonar_done)
  {
    /* fill rs->SonarController and all substructures with the
     * current values (and timestamps) */
    if ((error = N_GetSonar(robot_id)) != N_NO_ERROR)
    {
      printf("%s: retrieval of sonar values failed (%d).\n", argv0, error);
      exit(-1);
    }

    printf("---readings------------------------------"
	   "---------------------------------------\n");
    for (j=0; j < rs->SonarController.SonarSetCount; j++)
    {
      set = &(rs->SonarController.SonarSet[j]); 
      for (i=0; i < set->SonarCount; i++)
      {
	if (set->Sonar[i].Reading == N_SONAR_TIMEOUT) /* no reading */
	{
	  printf("       -- ");
	}
	else
	{
	  printf("%9ld ", set->Sonar[i].Reading);
	}
      }
      printf("\n");
    }

    printf("---time stamps --------------------------"
	   "---------------------------------------\n");
    for (j = 0; j < rs->SonarController.SonarSetCount; ++j)
    {
      set = &(rs->SonarController.SonarSet[j]); 
      for (i=0; i < set->SonarCount; i++)
      {
	printf("%9ld ", set->Sonar[i].TimeStamp);
      }
      printf("\n");
    }
    printf("\n");

    sleep(1);
  }

  /* restore the old sonar configuration */
  for (j = 0; j < rs->SonarController.SonarSetCount; ++j)
  {
    set = &rs->SonarController.SonarSet[j]; 
    set->FiringDelay = old_firing_delays[j];
    memcpy(set->FiringOrder, old_firing_orders[j], N_MAX_SONAR_COUNT + 1);
  }

  if ((error = N_SetSonarConfiguration(robot_id)) != N_NO_ERROR)
  {
    printf("%s: restoration of sonar configuration failed (%d).\n",
	   argv0, error);
    exit(-1);
  }

  N_DisconnectRobot(robot_id);
  return 0;
}
