/* xrdev/examples/swerve.c
 *
 * sonar based wandering with infrared based obstacle avoidance
 *
 *                             ** COPYRIGHT NOTICE **
 *
 * Copyright (c) 1999, Nomadic Technologies, Inc.
 *
 * The contents of this file is copyrighted by Nomadic Technologies, Inc, and
 * is protected by United States copyright laws.  Unauthorized copying is
 * prohibited.  All rights reserved.
 *
 */

#include <math.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>

#include "Nclient.h"

#include "robot_arg.h"



#define SONAR_TURN_DIST      900    /* start turning at this distance (mm) */
#define SONAR_TURN_GAIN      0.12   /* tweak this to change the turn rate */

#define SONAR_BOUNCE_DIST    200    /* bounce away at this distance (mm) */
#define SONAR_BOUNCE_GAIN    1.0    /* tweak this for stronger bounce */

#define IR_BOUNCE_THRESHOLD 20   /* the energy level that triggers avoidance */
#define IR_BOUNCE_GAIN      3.5  /* tweak this for stronger avoidance */

#define BUMPER_GAIN         500  /* speed at which we bounce using bumpers */

#define XY_SPEED      125           /* max translatation speed, in mm/sec */
#define XY_ACCEL      300           /* translational acceleration (mm/sec^2) */

#define PSI_DIST      2.0*M_PI      /* bound on random rotations */
#define PSI_SPEED     300           /* speed to rotate, in mrad/sec */
#define PSI_ACCEL     150           /* rotational accelearsion (mrad/sec^2) */
#define PSI_GAIN      0.5           /* damping constant */

/* modes for steering, set on the command line */
#define STEER_NONE    0             /* always face the same direction */
#define STEER_ORIGIN  1             /* always face +Y towards (0, 0) */
#define STEER_RANDOM  2             /* random rotations */

#define K             3.5


struct Vector
{
  float dir;
  float mag;
};



/* array of sensor indices in a circle.  first index is the set
 * number, second index is the sonar number.  we only describe the
 * top three sensor sets, as the bottom sets can be had by adding 1
 * to the set number */
static const int sw_sensor_circle[24][2] = 
{
  {4,6}, {4,7},
  {0,0}, {0,1}, {0,2}, {0,3}, {0,4}, {0,5}, {0,6}, {0,7},
  {2,0}, {2,1}, {2,2}, {2,3}, {2,4}, {2,5}, {2,6}, {2,7},
  {4,0}, {4,1}, {4,2}, {4,3}, {4,4}, {4,5} 
};

/* flag, set by the signal trapper to exit the main loop */
BOOL sw_done;




/* ------------------------------------------------------------------------
 * Function:     NormalizeAngle
 * Purpose:      convert an angle to [-PI, PI)
 * Arguments:    angle : the angle to be converted, in radians
 * Return:       static float : the normalized angle
 * ------------------------------------------------------------------------ */
static float NormalizeAngle(float angle)
{
  angle=fmod(angle, 2*M_PI);

  if (angle >= M_PI)
  {
    angle -= M_PI*2.0;
  }
  if (angle < -M_PI)
  {
    angle += M_PI*2.0;
  }
  return angle;
}



/* ------------------------------------------------------------------------
 * Function:     SonarReadingAtIndex
 * Purpose:      return the reading at a given sonar index in [0, 23]
 *                 corresponding to sonar positions about the robot.
 * Arguments:    *scp : from the state structure
 *               sn_index : the index to check
 * Return:       long : the minimum of the values from the top sonar ring
 *                 and the bottom sonar ring, at the given index
 * ------------------------------------------------------------------------ */
static long SonarReadingAtIndex(struct N_SonarController *scp, int sn_index)
{
  int set, sn;
  long r1, r2;

  set = sw_sensor_circle[sn_index][0];
  sn = sw_sensor_circle[sn_index][1];

  r1 = scp->SonarSet[set].Sonar[sn].Reading;
  r2 = scp->SonarSet[set + 1].Sonar[sn].Reading;

  return r1 < r2 ? r1 : r2;
}




/* ------------------------------------------------------------------------
 * Function:     IRReadingAtIndex
 * Purpose:      return the reading at a given infrared index in [0, 23]
 *                 corresponding to infrared positions about the robot.
 * Arguments:    *ircp : from the state structure
 *               ir_index : the index to check
 * Return:       long : the maximum of the values from the top infrared ring
 *                 and the bottom infrared ring, at the given index
 * ------------------------------------------------------------------------ */
static long IRReadingAtIndex(struct N_InfraredController *ircp, int ir_index)
{
  int set, in;
  long r1, r2;

  set = sw_sensor_circle[ir_index][0];
  in = sw_sensor_circle[ir_index][1];

  r1 = ircp->InfraredSet[set].Infrared[in].Reading;
  r2 = ircp->InfraredSet[set + 1].Infrared[in].Reading;

  return r1 > r2 ? r1 : r2;
}



/* ------------------------------------------------------------------------
 * Function:     BumperReadingAtIndex
 * Purpose:      return the reading at a given bumper index in [0, 23]
 *                 corresponding to bumper positions about the robot.
 * Arguments:    *bcp : from the state structure
 *               bmp_index : the index to check
 * Return:       char : the maximum of the values from the top bumper ring
 *                 and the bottom bumper ring, at the given index
 * ------------------------------------------------------------------------ */
static char BumperReadingAtIndex(struct N_BumperController *bcp, int bmp_index)
{
  int set, bn;
  char r1, r2;

  set = sw_sensor_circle[bmp_index][0];
  bn = sw_sensor_circle[bmp_index][1];

  r1 = bcp->BumperSet[set].Bumper[bn].Reading;
  r2 = bcp->BumperSet[set + 1].Bumper[bn].Reading;

  return r1 > r2 ? r1 : r2;
}



/* ------------------------------------------------------------------------
 * Function:     GetSonarVector
 * Purpose:      modify a given vector with a "sonar swerve wandering"
 *                 behavior
 * Arguments:    struct N_SonarController *scp : from the state structure
 *               struct Vector *vp : the current direction of travel
 * ------------------------------------------------------------------------ */
static void GetSonarVector(struct N_SonarController *scp, struct Vector *vp)
{
  int i, sn, count = 0;
  float angle, x = 0.0, y = 0.0;
  long reading;
  struct Vector obstacle;

  /* find the index in [0, 23] which corresponds to the current direction */
  sn = (int)(NormalizeAngle(vp->dir)/2.0/M_PI*24.0 + 18.5) % 24;
  for (i = 0; i < 12; ++i, sn = ++sn % 24)
  {
    reading = SonarReadingAtIndex(scp, sn);
    if (reading < SONAR_TURN_DIST)
    {
      angle = sn*M_PI*2.0/24.0;
      x += (SONAR_TURN_DIST - reading)*cos(angle);
      y += (SONAR_TURN_DIST - reading)*sin(angle);
      ++count;
    }
  }

  if (count == 0)
  {
    return;
  }

  /* swerve at 90 degrees to the obstacle, or go in the current
   * direction, whichever is closer */
  obstacle.dir = atan2(y, x) - M_PI;
  obstacle.mag = SONAR_TURN_GAIN*hypot(x, y)/(float)count;
  angle = NormalizeAngle(vp->dir - obstacle.dir);
  if (angle > M_PI/2.0)
  {
    angle = M_PI/2.0 - 0.3;
  }
  if (angle < -M_PI/2.0)
  {
    angle = -M_PI/2.0 + 0.3;
  }
  obstacle.dir = NormalizeAngle(obstacle.dir + angle);

  x = obstacle.mag*cos(obstacle.dir) + vp->mag*cos(vp->dir);
  y = obstacle.mag*sin(obstacle.dir) + vp->mag*sin(vp->dir);
  vp->dir = atan2(y, x);
  vp->mag = hypot(x, y);
  if (vp->mag > XY_SPEED)
  {
    vp->mag = XY_SPEED;
  }

  return;
}



/* ------------------------------------------------------------------------
 * Function:     GetIRVector
 * Purpose:      modify the given motion vector so that we avoid any
 *                 obstacles detected by the infrareds 
 * Arguments:    *ircp : from the state structure
 *               *vp : the current direction of travel
 * ------------------------------------------------------------------------ */
static void GetIRVector(struct N_InfraredController *ircp, struct Vector *vp)
{
  int i, in, count = 0;
  float x, y, angle;
  long reading;
  struct Vector obstacle;

  /* find the index in [0, 23] which corresponds to the current direction */
  in = (int)(NormalizeAngle(vp->dir)/2.0/M_PI*24.0 + 18.5) % 24;
  for (i = 0; i < 12; ++i, in = ++in % 24)
  {
    reading = IRReadingAtIndex(ircp, in);
 
    if (reading > IR_BOUNCE_THRESHOLD)
    {
      angle = in*M_PI*2.0/24.0;
      x += (reading - IR_BOUNCE_THRESHOLD)*cos(angle);
      y += (reading - IR_BOUNCE_THRESHOLD)*sin(angle);
      ++count;
    }
  }

  if (count == 0)
  {
    return;
  }

  /* bounce directly away from close obstacles */
  obstacle.dir = NormalizeAngle(atan2(y, x) - M_PI);
  obstacle.mag = IR_BOUNCE_GAIN*hypot(x, y)/(float)count;

  x = obstacle.mag*cos(obstacle.dir) + vp->mag*cos(vp->dir);
  y = obstacle.mag*sin(obstacle.dir) + vp->mag*sin(vp->dir);
  vp->dir = atan2(y, x);
  vp->mag = hypot(x, y);
  if (vp->mag > XY_SPEED)
  {
    vp->mag = XY_SPEED;
  }

  return;
}



static void GetBumperVector(struct N_BumperController *bcp, struct Vector *vp)
{
  int i, bn, count = 0;
  float angle, x = 0.0, y = 0.0;
  char reading;

  /* find the index in [0, 23] which corresponds to the current direction */
  bn = (int)(NormalizeAngle(vp->dir)/2.0/M_PI*24.0 + 18.5) % 24;
  for (i = 0; i < 12; ++i, bn = ++bn % 24)
  {
    reading = BumperReadingAtIndex(bcp, bn);
      
    if (reading == N_BUMPER_HIGH)
    {
      angle = bn*M_PI*2.0/24.0;
      x += reading*cos(angle);
      y += reading*sin(angle);
      ++count;
    }
  }
  
  if (count == 0)
  {
    return;
  }

  /* ignore current goal, and make it go away from the obstacle */
  vp->dir = NormalizeAngle(atan2(y, x) - M_PI);
  vp->mag = BUMPER_GAIN*hypot(x, y)/(float)count;
  printf("bumper hits, moving to %f @ %f\n", vp->mag, vp->dir);
  return;
}
      



/* ------------------------------------------------------------------------
 * Function:     SteerToOrigin
 * Purpose:      modify the robot's rotational destination so that we
 *                 always face towards the origin (where the robot was
 *                 zeroed)
 * Arguments:    struct N_RobotState *rs : the state structure
 * ------------------------------------------------------------------------ */
static void SteerToOrigin(struct N_RobotState *rs)
{
  static float last_psi = 0.0;
  float psi, new_psi, x, y, dx, dy;

  /* we compute the (x, y) vector to the destination and the robot's current
   * rotation angle.  we then rotate the destination vector through the
   * negative angle and compute the angle from 0 degrees to the destination.
   * this gives us an angle in [-PI, PI] which is the shortest rotational
   * distance to the direction we want to face */
  dx = -rs->AxisSet.Axis[N_XTRANSLATION].ActualPosition;
  dy = -rs->AxisSet.Axis[N_YTRANSLATION].ActualPosition;
  psi = rs->AxisSet.Axis[N_ROTATION].ActualPosition/1000.0;

  x = cos(psi)*dx + sin(psi)*dy;
  y = cos(psi)*dy - sin(psi)*dx;
  new_psi = atan2(y, x);

  /* PSI_GAIN is a damping constant to prevent jerky motions */
  last_psi = last_psi*PSI_GAIN + (psi + atan2(y, x))*(1.0 - PSI_GAIN);

  rs->AxisSet.Axis[N_ROTATION].Mode = N_AXIS_POSITION_ABSOLUTE;
  rs->AxisSet.Axis[N_ROTATION].DesiredPosition = last_psi*1000;
  rs->AxisSet.Axis[N_ROTATION].DesiredSpeed = PSI_SPEED;
  rs->AxisSet.Axis[N_ROTATION].Acceleration = PSI_ACCEL;
  rs->AxisSet.Axis[N_ROTATION].Update = TRUE;

  return;
}




/* ------------------------------------------------------------------------
 * Function:     SteerRandom
 * Purpose:      modify the robot's rotational destination to a random
 *                 bounded goal, if the current motion has completed
 * Arguments:    struct N_RobotState *rs : the state structure
 * ------------------------------------------------------------------------ */
static void SteerRandom(struct N_RobotState *rs)
{
  if (rs->AxisSet.Axis[N_ROTATION].InProgress)
  {
    return;
  }

  rs->AxisSet.Axis[N_ROTATION].Mode = N_AXIS_POSITION_ABSOLUTE;
  rs->AxisSet.Axis[N_ROTATION].DesiredPosition = 
    (((double)PSI_DIST*rand())/(RAND_MAX + 1.0) - PSI_DIST/2.0)*1000;
  rs->AxisSet.Axis[N_ROTATION].DesiredSpeed = PSI_SPEED;
  rs->AxisSet.Axis[N_ROTATION].Acceleration = PSI_ACCEL;
  rs->AxisSet.Axis[N_ROTATION].Update = TRUE;
  
  return;
}




/* ------------------------------------------------------------------------
 * Function:     SignalTrap
 * Purpose:      catch process interrupts, so that we can exit the main
 *                 loop and clean up after ourselves
 * ------------------------------------------------------------------------ */
static void SignalTrap(int signal_num)
{
  sw_done = TRUE;
  return;
}



int main(int argc, char *argv[])
{
  BOOL axes_need_reset;
  char *sched_host;
  unsigned short sched_port;
  int error, i, j, steer_mode;
  long robot_id;
  struct N_RobotState *rs;
  struct N_SonarSet *set;
  struct Vector vel = { M_PI/2.0, XY_SPEED }; /* initial direction is +Y */
  struct Vector lastVel = {0.0, 0.0};
  struct timeval tv0, tv1;
  float dt = 0.0;

  /* get the robot ID and scheduler from the command line */
  sched_host = "localhost";
  sched_port = 7073;
  robot_id = 1;
  ARG_GetStandardArgs(&argc, &argv, &sched_host, &sched_port, &robot_id);

  i = 1;
  steer_mode = STEER_RANDOM;
  while (i < argc)
  {
    if (strcasecmp(argv[i], "--origin") == 0)
    {
      steer_mode = STEER_ORIGIN;
    }
    else if (strcasecmp(argv[i], "--nosteer") == 0)
    {
      steer_mode = STEER_NONE;
    }
  }

  if ((error = N_InitializeClient(sched_host, sched_port)) != N_NO_ERROR ||
      (error = N_ConnectRobot(robot_id)) != N_NO_ERROR)
  {
    fprintf(stderr, "%s: error connecting (%d).\n", argv[0], error);
    exit(-1);
  }

  rs = N_GetRobotState(robot_id);

  /* here, we zero the robot.  before setting the integrated configuration
   * to zero, we must make sure that none of the axes have a non-zero
   * goal position so that the zeroing does not induce a move */
  rs->AxisSet.Global = TRUE;
  rs->AxisSet.Axis[N_XTRANSLATION].Mode = N_AXIS_POSITION_ABSOLUTE;
  rs->AxisSet.Axis[N_XTRANSLATION].DesiredPosition = 0;
  rs->AxisSet.Axis[N_XTRANSLATION].DesiredSpeed = 0;
  rs->AxisSet.Axis[N_XTRANSLATION].Acceleration = 0;
  rs->AxisSet.Axis[N_XTRANSLATION].Update = TRUE;
  rs->AxisSet.Axis[N_YTRANSLATION].Mode = N_AXIS_POSITION_ABSOLUTE;
  rs->AxisSet.Axis[N_YTRANSLATION].DesiredPosition = 0;
  rs->AxisSet.Axis[N_YTRANSLATION].DesiredSpeed = 0;
  rs->AxisSet.Axis[N_YTRANSLATION].Acceleration = 0;
  rs->AxisSet.Axis[N_YTRANSLATION].Update = TRUE;
  rs->AxisSet.Axis[N_ROTATION].Mode = N_AXIS_POSITION_ABSOLUTE;
  rs->AxisSet.Axis[N_ROTATION].DesiredPosition = 0;
  rs->AxisSet.Axis[N_ROTATION].DesiredSpeed = 0;
  rs->AxisSet.Axis[N_ROTATION].Acceleration = 0;
  rs->AxisSet.Axis[N_ROTATION].Update = TRUE;
  if ((error = N_SetAxes(robot_id)) != N_NO_ERROR)
  {
    fprintf(stderr, "%s: error setting axes to zero (%d).\n", argv[0], error);
    exit(-1);
  }

  /* zero the integrated configuration */
  rs->Integrator.x = 0;
  rs->Integrator.y = 0;
  rs->Integrator.Rotation = 0;
  if ((error = N_SetIntegratedConfiguration(robot_id)) != N_NO_ERROR)
  {
    fprintf(stderr, "%s: error zeroing (%d).\n", argv[0], error);
    exit(-1);
  }

  /* configure the sonars for full firing order */
  for (j = 0; j < rs->SonarController.SonarSetCount; j++)
  {
    set = &rs->SonarController.SonarSet[j]; 
    set->FiringDelay = 0;
    set->TimeStampActive = TRUE;
    for (i = 0; i < set->SonarCount; i++)
    {
      set->FiringOrder[i] = i;
    }
  }
  if ((error = N_SetSonarConfiguration(robot_id)) != N_NO_ERROR)
  {
    fprintf(stderr, "%s: error initializing sonars (%d).\n", argv[0], error);
    exit(-1);
  }

  /* loop until SIGINT */
  sw_done = FALSE;
  signal(SIGINT, SignalTrap);
  while (!sw_done)
  {
    gettimeofday(&tv0, NULL);

    if ((error = N_GetAxes(robot_id)) != N_NO_ERROR)
    {
      fprintf(stderr, "error retrieving axis values (%d).\n", error);
      exit(-1);
    }

    /* check for estop or joystick */
    if (rs->AxisSet.Status != N_AXES_READY)
    {
      if (!axes_need_reset)
      {
	axes_need_reset = TRUE;
      }
      continue;
    }

    if (axes_need_reset)
    {
      /* the joystick or another client may have unset global mode.  we reset
       * it here to make sure that N_GetAxes returns the appropriate values */
      rs->AxisSet.Global = TRUE;
      if ((error = N_SetAxes(robot_id)) != N_NO_ERROR &&
	  error != N_AXES_NOT_READY)
      {
	fprintf(stderr, "error resetting axes (%d).\n", error);
	exit(-1);
      }
      axes_need_reset = FALSE;
      continue;
    }

    /* steer the robot according to the user's preferences */
    switch (steer_mode)
    {
    case STEER_ORIGIN: 
      SteerToOrigin(rs);
      break;

    case STEER_RANDOM:
      SteerRandom(rs);
      break;
    }

    /* assume we can go full speed */
    vel.mag = XY_SPEED;

    /* modify the current direction based on the sonar wandering behavior */
    if ((error = N_GetSonar(robot_id)) != N_NO_ERROR)
    {
      fprintf(stderr, "error retrieving sonar values (%d).\n", error);
      exit(-1);
    }
    GetSonarVector(&rs->SonarController, &vel);

    /* modify using the infrared based avoidance behavior */
    if ((error = N_GetInfrared(robot_id)) != N_NO_ERROR)
    {
      fprintf(stderr, "error retrieving infrared values (%d).\n", error);
      exit(-1);
    }
    GetIRVector(&rs->InfraredController, &vel);

    if ((error = N_GetBumper(robot_id)) != N_NO_ERROR)
    {
      fprintf(stderr, "error retrieving bumper values (%d).\n", error);
      exit(-1);
    }
    GetBumperVector(&rs->BumperController, &vel);

    /* change the vel vector based on time:  
     *
     * Our time-varying motion vector (v) is described by:
     *
     * v(t)=v(t - dt) + vsensor1(t) + vsensor2(t) + ... + vsensorn(t)
     * v(t)=v(t - dt) + vsensors(t)
     *
     * Thus, v(t) gets updated at 1/dt times per second.  For constant 
     * vsensors(t), v(t) will change by 1/dt*vsensors(t) every second.  
     * For small dt, this can be a big change.
     *
     * To normalize with respect to time we multiply vsensors(t) by dt.
     *
     * To obtain vsensors using above expression:
     *      
     * vsensors(t) = v(t) - v(t - dt)
     *
     * Thus for normalized motion vector (vnorm):
     *
     * vnorm(t)= v(t - dt) + K * dt * (v(t) - v(t - dt))
     **   
     * where K is our tuning constant that will give us the correct response.
     * Assume 1/dt = 3.5 for prior correct response.  This implies K should 
     * be about 3.5. */
    
    vel.mag = lastVel.mag + K*dt*(vel.mag - lastVel.mag);
    vel.dir = lastVel.dir + K*dt*(vel.dir - lastVel.dir);
    vel.dir = NormalizeAngle(vel.dir);

    /* set up the new motion, given the velocity vector */
    rs->AxisSet.Global = TRUE;
    rs->AxisSet.Axis[N_XTRANSLATION].Mode = N_AXIS_VELOCITY;
    rs->AxisSet.Axis[N_XTRANSLATION].DesiredSpeed = vel.mag*
      cos(vel.dir + rs->AxisSet.Axis[N_ROTATION].ActualPosition/1000.0);
    rs->AxisSet.Axis[N_XTRANSLATION].Acceleration = XY_ACCEL;
    rs->AxisSet.Axis[N_XTRANSLATION].Update = TRUE;
    rs->AxisSet.Axis[N_YTRANSLATION].Mode = N_AXIS_VELOCITY;
    rs->AxisSet.Axis[N_YTRANSLATION].DesiredSpeed = vel.mag*
      sin(vel.dir + rs->AxisSet.Axis[N_ROTATION].ActualPosition/1000.0);
    rs->AxisSet.Axis[N_YTRANSLATION].Acceleration = XY_ACCEL;
    rs->AxisSet.Axis[N_YTRANSLATION].Update = TRUE;

    /* do the motion.  if the axes aren't ready, this probably means
     * that the joystick was pressed; this will be checked in N_GetAxes
     * above. */
    if ((error = N_SetAxes(robot_id)) != N_NO_ERROR &&
	error != N_AXES_NOT_READY)
    {
      fprintf(stderr, "error sending motion comand (%d).\n", error);
      exit(-1);
    }

    gettimeofday(&tv1, NULL);
    
    dt = (float)((tv1.tv_sec - tv0.tv_sec)*1000000 + 
		 tv1.tv_usec - tv0.tv_usec)/1000000.0;

    lastVel = vel;
  } /* while not done */

  /* stop the robot */
  rs->AxisSet.Global = TRUE;
  rs->AxisSet.Axis[N_XTRANSLATION].Mode = N_AXIS_STOP;
  rs->AxisSet.Axis[N_YTRANSLATION].Mode = N_AXIS_STOP;
  rs->AxisSet.Axis[N_ROTATION].Mode = N_AXIS_STOP;
  rs->AxisSet.Axis[N_XTRANSLATION].Update = TRUE;
  rs->AxisSet.Axis[N_YTRANSLATION].Update = TRUE;
  rs->AxisSet.Axis[N_ROTATION].Update = TRUE;
  if ((error = N_SetAxes(robot_id)) != N_NO_ERROR)
  {
    fprintf(stderr, "%s: error stopping the robot (%d).\n", argv[0], error);
    exit(-1);
  }

  N_DisconnectRobot(robot_id);
  return 0;
}
